/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "client.h"
#include "stub_manager.h"
#include <google/protobuf/util/json_util.h>
#include <grpcpp/grpcpp.h>
#include <iostream>
#include <unistd.h>
#include <nlohmann/json.hpp>
#include <sys/socket.h>
using grpc::ClientContext;
using grpc::Status;
using json = nlohmann::json;
using grpc::ClientAsyncResponseReaderInterface;
#define LOG_INFO(msg)                                                          \
  {                                                                            \
    time_t t = time(0);                                                        \
    char mbstr[100];                                                           \
    strftime(mbstr, sizeof(mbstr), "%Y-%m-%d %X", localtime(&t));              \
    std::cout << mbstr << " - INFO - " << msg << std::endl;                    \
  }
#define LOG_ERROR(msg)                                                         \
  {                                                                            \
    time_t t = time(0);                                                        \
    char mbstr[100];                                                           \
    strftime(mbstr, sizeof(mbstr), "%Y-%m-%d %X", localtime(&t));              \
    std::cerr << mbstr << " - ERROR - " << msg << std::endl;                   \
  }

namespace document {

std::string ProtoMessageToJsonString(const google::protobuf::Message &message);
template <typename T>
T JsonStringToProtoMessage(const std::string &json_string);
std::string ProtoMessageToJsonString(const google::protobuf::Message &message) {
  std::string json_string;
  google::protobuf::util::JsonPrintOptions options;
  options.add_whitespace = true;
  options.always_print_primitive_fields = true;
  google::protobuf::util::MessageToJsonString(message, &json_string, options);
  return json_string;
}
template <typename T>
T JsonStringToProtoMessage(const std::string &json_string) {
  T message;
  google::protobuf::util::JsonParseOptions options;
  options.ignore_unknown_fields = true;
  google::protobuf::util::JsonStringToMessage(json_string, &message, options);
  return message;
}

static std::string ConstructSockPath(const std::string& base_path, const std::string& suffix) {
    uid_t uid = getuid(); // 获取当前用户的UID
    return base_path + std::to_string(uid) + suffix;
}

std::shared_ptr<grpc::Channel> CreateChannelWithMaxMessageSize(const std::string &target) {
    grpc::ChannelArguments args;
    args.SetMaxReceiveMessageSize(1024 * 1024 * 1024);
    return grpc::CreateCustomChannel(target, grpc::InsecureChannelCredentials(), args);
}

document::DocumentChunkClient::DocumentChunkClient()
    : stub_manager_(std::make_unique<ChunkStubManager>(
          CreateChannelWithMaxMessageSize(
              ConstructSockPath("unix:///tmp/document_service_sync_", ".sock")),
          CreateChannelWithMaxMessageSize(ConstructSockPath(
              "unix:///tmp/document_service_async_", ".sock")))) {}

DocumentChunkClient::~DocumentChunkClient() {}

DocumentChunkResponse DocumentChunkClient::SyncChunkDocument(
    const std::vector<std::string> &file_paths,
    const std::vector<std::string> &chunker_types,
    const std::vector<int32_t> &chunk_sizes) {
  DocumentChunkResponse response_json;
  ChunkRequest request;
  for (size_t i = 0; i < file_paths.size(); ++i) {
    request.add_file_paths(file_paths[i]);
    request.add_chunker_types(chunker_types[i]);
    request.add_chunk_sizes(chunk_sizes[i]);
  }
  std::string request_json = ProtoMessageToJsonString(request);
  LOG_INFO("request: " + request_json);
  ChunkedResponse response;
  ClientContext context;
  Status status = stub_manager_->GetSyncStub()->SyncChunkDocument(
      &context, request, &response);
  std::string response_json_str = ProtoMessageToJsonString(response);
  LOG_INFO("Chunked over: " + request_json);
  response_json.data = response_json_str;
  response_json.errorCode = status.error_code();
  response_json.errorMsg = status.error_message();
  return response_json;
}

void DocumentChunkClient::AsyncChunkDocument(
    const std::vector<std::string> &file_paths,
    const std::vector<std::string> &chunker_types,
    const std::vector<int32_t> &chunk_sizes, 
    void *callback_data,
    DocumentChunkCallback callback) {
  ChunkRequest request;
  ChunkedResponse response;
  for (size_t i = 0; i < file_paths.size(); ++i) {
    request.add_file_paths(file_paths[i]);
    request.add_chunker_types(chunker_types[i]);
    request.add_chunk_sizes(chunk_sizes[i]);
  }
  std::string request_json = ProtoMessageToJsonString(request);
  LOG_INFO("request: " + request_json);
  ClientContext context;
  auto *async_stub = stub_manager_->GetAsyncStub();
  auto wrapped_callback = [this, callback, callback_data](::grpc::Status status) {
    ChunkedResponse response;
    DocumentChunkResponse response_json;
    if (status.ok()) {
      std::string response_json_str = ProtoMessageToJsonString(response);
      response_json.data = response_json_str;
      response_json.errorCode = status.error_code();
      response_json.errorMsg = status.error_message();
    } else {
      LOG_ERROR("Async request failed: " + status.error_message());
      response_json.data = nullptr;
      response_json.errorCode = status.error_code();
      response_json.errorMsg = status.error_message();
    }
    callback(response_json, callback_data);
  };
  async_stub->experimental_async()->AsyncChunkDocument(&context, &request, &response, wrapped_callback);
}
} // namespace document